@@ -30,6 +30,7 @@ from nuitka import SourceCodeReferences
 from nuitka.__past__ import iter_modules
 from nuitka.containers.Namedtuples import makeNamedtupleClass
 from nuitka.containers.OrderedSets import OrderedSet
+from nuitka.Errors import NuitkaCodeDeficit
 from nuitka.plugins.Plugins import Plugins
 from nuitka.PythonFlavors import isNuitkaPython
 from nuitka.PythonVersions import python_version
@@ -37,11 +38,11 @@ from nuitka.Tracing import recursion_logger
 from nuitka.tree.ReformulationMultidist import locateMultidistModule
 from nuitka.utils.AppDirs import getCacheDir
 from nuitka.utils.FileOperations import (
+    getFileContentsHash,
     getNormalizedPath,
     listDir,
     removeDirectory,
 )
-from nuitka.utils.Hashing import getFileContentsHash
 from nuitka.utils.Importing import (
     builtin_module_names,
     getExtensionModuleSuffixes,
@@ -66,12 +67,12 @@ from .StandardLibrary import isStandardLibraryPath
 # Debug traces, enabled via --explain-imports
 _debug_module_finding = None
 
-# Preference as expressed via --prefer-source-code
-_prefer_source_code_over_extension_modules = None
-
 # Do not add current directory to search path.
 _safe_path = None
 
+# Debug mode
+_is_debug = None
+
 
 def setupImportingFromOptions():
     """Set up the importing layer from giving options."""
@@ -83,16 +84,14 @@ def setupImportingFromOptions():
     global _debug_module_finding
     _debug_module_finding = Options.shallExplainImports()
 
-    global _prefer_source_code_over_extension_modules
-    _prefer_source_code_over_extension_modules = (
-        Options.shallPreferSourceCodeOverExtensionModules()
-    )
-
     global _safe_path
     _safe_path = Options.hasPythonFlagNoCurrentDirectoryInPath()
 
+    global _is_debug
+    _is_debug = Options.is_debug
+
     # Lets try and have this complete, please report failures.
-    if Options.is_debug and not isNuitkaPython():
+    if _is_debug and not isNuitkaPython():
         _checkRaisingBuiltinComplete()
 
     main_filenames = Options.getMainEntryPointFilenames()
@@ -134,6 +133,9 @@ def makeModuleUsageAttempt(
 ):
     assert source_ref is not None
 
+    # The looked for module usage attempt, cannot be a relative module name
+    assert module_name.getTopLevelPackageName() != "", source_ref
+
     return ModuleUsageAttempt(
         module_name=module_name,
         filename=filename,
@@ -361,6 +363,10 @@ def normalizePackageName(module_name):
     return module_name
 
 
+# Avoid creating this over and over
+_find_module_not_found = (None, None, None, None, "not-found")
+
+
 def findModule(module_name, parent_package, level, logger):
     """Find a module with given package name as parent.
 
@@ -393,12 +399,12 @@ def findModule(module_name, parent_package, level, logger):
         if parent_package is not None:
             parent_package = parent_package.getRelativePackageName(level - 1)
         else:
-            return None, None, None, "not-found"
+            return _find_module_not_found
 
     if level == 1 and not module_name:
         # Not actually allowed, but we only catch that at run-time.
         if parent_package is None:
-            return None, None, None, "not-found"
+            return _find_module_not_found
 
         module_name = parent_package
         parent_package = None
@@ -422,10 +428,10 @@ def findModule(module_name, parent_package, level, logger):
             else:
                 module_filename = None
 
-            return full_name.getPackageName(), module_filename, "py", "pth"
+            return full_name, full_name.getPackageName(), module_filename, "py", "pth"
 
         try:
-            module_filename, module_kind = _findModule(
+            found_module_name, module_filename, module_kind = _findModule(
                 module_name=full_name,
                 logger=logger,
             )
@@ -439,7 +445,13 @@ def findModule(module_name, parent_package, level, logger):
                     % (module_name, full_name, module_filename)
                 )
 
-            return full_name.getPackageName(), module_filename, module_kind, "relative"
+            return (
+                found_module_name,
+                full_name.getPackageName(),
+                module_filename,
+                module_kind,
+                "relative",
+            )
 
     if level < 1 and module_name:
         module_name = normalizePackageName(module_name)
@@ -455,10 +467,10 @@ def findModule(module_name, parent_package, level, logger):
             else:
                 module_filename = None
 
-            return package_name, module_filename, "py", "pth"
+            return module_name, package_name, module_filename, "py", "pth"
 
         try:
-            module_filename, module_kind = _findModule(
+            found_module_name, module_filename, module_kind = _findModule(
                 module_name=module_name,
                 logger=logger,
             )
@@ -469,12 +481,18 @@ def findModule(module_name, parent_package, level, logger):
             if logger is not None:
                 logger.info(
                     "findModule: Found absolute imported module '%s' in filename '%s':"
-                    % (module_name, module_filename)
+                    % (found_module_name, module_filename)
                 )
 
-            return package_name, module_filename, module_kind, "absolute"
+            return (
+                found_module_name,
+                package_name,
+                module_filename,
+                module_kind,
+                "absolute",
+            )
 
-    return None, None, None, "not-found"
+    return _find_module_not_found
 
 
 # Some platforms are case insensitive.
@@ -482,39 +500,9 @@ case_sensitive = not isMacOS() and not isWin32OrPosixWindows()
 
 ImportScanFinding = collections.namedtuple(
     "ImportScanFinding",
-    ("found_in", "module_type", "priority", "full_path", "search_order"),
+    ("found_as", "found_in", "module_type", "priority", "full_path", "search_order"),
 )
 
-# We put here things that are not worth it (Cython is not really used by
-# anything really, or where it's know to not have a big # impact, e.g. lxml.
-
-# TODO: Do we really want this warning to persist?
-unworthy_namespaces = ("Cython", "lxml", "black", "tomli")
-
-
-def _reportCandidates(package_name, module_name, candidate, candidates):
-    module_name = (
-        package_name.getChildNamed(module_name)
-        if package_name is not None
-        else module_name
-    )
-
-    if candidate.priority == 1 and _prefer_source_code_over_extension_modules is None:
-        for c in candidates:
-            # Don't compare to itself and don't consider unused bytecode a problem.
-            if c is candidate or c.priority == 3:
-                continue
-
-            if c.search_order == candidate.search_order:
-                if not module_name.hasOneOfNamespaces(unworthy_namespaces):
-                    recursion_logger.info(
-                        """\
-Should decide '--prefer-source-code' vs. '--no-prefer-source-code', using \
-existing '%s' extension module by default. Candidates were: %s <-> %s."""
-                        % (module_name, candidate, c)
-                    )
-
-
 _list_dir_cache = {}
 
 
@@ -538,6 +526,105 @@ def flushImportCache():
     module_search_cache.clear()
 
 
+def _findModuleInPath3(
+    module_name,
+    package_name,
+    search_path,
+    search_index,
+    search_path_entry,
+    package_directory,
+):
+    # First, check for a package with an init file, that would be the
+    # first choice.
+
+    # Higher values are lower priority.
+    priority_map = {
+        "PY_COMPILED": 3,
+        "PY_SOURCE": 2,
+        "C_EXTENSION": 1,
+    }
+
+    if os.path.isdir(package_directory):
+        found = False
+
+        for suffix, module_type in getModuleFilenameSuffixes():
+            package_file_name = "__init__" + suffix
+
+            file_path = os.path.join(package_directory, package_file_name)
+
+            if os.path.isfile(file_path):
+                yield (
+                    ImportScanFinding(
+                        found_as=ModuleName.makeModuleNameInPackage(
+                            module_name, package_name
+                        ),
+                        found_in=search_path_entry,
+                        module_type=module_type,
+                        priority=priority_map[module_type],
+                        full_path=package_directory,
+                        search_order=search_index,
+                    )
+                )
+                found = True
+
+        if not found and python_version >= 0x300:
+            yield (
+                ImportScanFinding(
+                    found_as=ModuleName.makeModuleNameInPackage(
+                        module_name, package_name
+                    ),
+                    found_in=search_path_entry,
+                    module_type=10,
+                    priority=10,
+                    full_path=package_directory,
+                    search_order=search_index + len(search_path),
+                )
+            )
+
+    # Then, check out suffixes of all kinds, but only for one directory.
+    if search_path_entry is not None:
+        last_module_type = 0
+        for suffix, module_type in getModuleFilenameSuffixes():
+            # Use first match per kind only.
+            if module_type == last_module_type:
+                continue
+
+            full_path = os.path.join(search_path_entry, module_name + suffix)
+
+            if os.path.isfile(full_path):
+                yield (
+                    ImportScanFinding(
+                        found_as=ModuleName.makeModuleNameInPackage(
+                            module_name, package_name
+                        ),
+                        found_in=search_path_entry,
+                        module_type=module_type,
+                        priority=4 + priority_map[module_type],
+                        full_path=full_path,
+                        search_order=search_index,
+                    )
+                )
+                last_module_type = module_type
+
+
+def _getSetuptoolsDistutilsPackageDir():
+    setuptools_package_dir = locateModule(
+        ModuleName("setuptools"), parent_package=None, level=0
+    )[1]
+
+    if setuptools_package_dir is not None:
+        setuptools_package_dir = os.path.join(setuptools_package_dir, "_distutils")
+
+    return setuptools_package_dir
+
+
+_recompile_extension_modules = {}
+
+
+def getRecompileDecisionReason(module_name):
+    return _recompile_extension_modules.get(module_name, (None, None))
+
+
 def _findModuleInPath2(package_name, module_name, search_path, logger):
     """This is out own module finding low level implementation.
 
@@ -546,7 +633,7 @@ def _findModuleInPath2(package_name, module_name, search_path, logger):
     None, if it is a built-in.
     """
     # We have many branches here, because there are a lot of cases to try.
-    # pylint: disable=too-many-branches,too-many-locals
+    # pylint: disable=too-many-branches
 
     # We may have to decide between package and module, therefore build
     # a list of candidates.
@@ -554,75 +641,39 @@ def _findModuleInPath2(package_name, module_name, search_path, logger):
 
     considered = set()
 
-    # Higher values are lower priority.
-    priority_map = {
-        "PY_COMPILED": 3,
-        "PY_SOURCE": 0 if _prefer_source_code_over_extension_modules else 2,
-        "C_EXTENSION": 1,
-    }
-
-    for count, entry in enumerate(search_path):
+    for count, search_path_entry in enumerate(search_path):
         # Don't try again, just with an entry of different casing or complete
         # duplicate.
-        if os.path.normcase(entry) in considered:
+        if os.path.normcase(search_path_entry) in considered:
             continue
-        considered.add(os.path.normcase(entry))
+        considered.add(os.path.normcase(search_path_entry))
 
-        package_directory = os.path.join(entry, module_name.asPath())
+        package_directory = os.path.join(search_path_entry, module_name.asPath())
 
-        # First, check for a package with an init file, that would be the
-        # first choice.
-        if os.path.isdir(package_directory):
-            found = False
+        for candidate in _findModuleInPath3(
+            module_name=module_name,
+            package_name=package_name,
+            search_path=search_path,
+            search_index=count,
+            search_path_entry=search_path_entry,
+            package_directory=package_directory,
+        ):
+            candidates.add(candidate)
 
-            for suffix, module_type in getModuleFilenameSuffixes():
-                package_file_name = "__init__" + suffix
+    # TODO: Make this a plugin decision instead.
+    if not candidates and package_name is None and module_name == "distutils":
+        setuptools_package_dir = _getSetuptoolsDistutilsPackageDir()
 
-                file_path = os.path.join(package_directory, package_file_name)
-
-                if os.path.isfile(file_path):
-                    candidates.add(
-                        ImportScanFinding(
-                            found_in=entry,
-                            module_type=module_type,
-                            priority=priority_map[module_type],
-                            full_path=package_directory,
-                            search_order=count,
-                        )
-                    )
-                    found = True
-
-            if not found and python_version >= 0x300:
-                candidates.add(
-                    ImportScanFinding(
-                        found_in=entry,
-                        module_type=10,
-                        priority=10,
-                        full_path=package_directory,
-                        search_order=count + len(search_path),
-                    )
-                )
-
-        # Then, check out suffixes of all kinds, but only for one directory.
-        last_module_type = 0
-        for suffix, module_type in getModuleFilenameSuffixes():
-            # Use first match per kind only.
-            if module_type == last_module_type:
-                continue
-
-            full_path = os.path.join(entry, module_name + suffix)
-
-            if os.path.isfile(full_path):
-                candidates.add(
-                    ImportScanFinding(
-                        found_in=entry,
-                        module_type=module_type,
-                        priority=4 + priority_map[module_type],
-                        full_path=full_path,
-                        search_order=count,
-                    )
-                )
-                last_module_type = module_type
+        if setuptools_package_dir is not None:
+            for candidate in _findModuleInPath3(
+                module_name=module_name,
+                package_name=None,
+                search_path=[],
+                search_path_entry=None,
+                package_directory=setuptools_package_dir,
+                search_index=-1,
+            ):
+                candidates.add(candidate)
 
     if logger is not None:
         logger.info("Candidates: %r" % candidates)
@@ -630,20 +681,43 @@ def _findModuleInPath2(package_name, module_name, search_path, logger):
     found_candidate = None
 
     if candidates:
-        # Sort by priority, with entries from same path element coming first, then desired type.
-        candidates = tuple(
-            sorted(candidates, key=lambda c: (c.search_order, c.priority))
-        )
+        # Sort by priority, with entries from same path element coming first, then desired type. In case
+        # pf there being an extension module,
+        if any(c.module_type == "C_EXTENSION" for c in candidates):
+
+            def prioritize(c):
+                if c.module_type == "PY_SOURCE":
+                    decision, reason = Plugins.decideRecompileExtensionModules(
+                        c.found_as
+                    )
+                    _recompile_extension_modules[c.found_as] = decision, reason
+                else:
+                    decision = False
+
+                return (
+                    c.search_order,
+                    ((c.priority - 2) if decision else c.priority),
+                )
+
+        else:
+
+            def prioritize(c):
+                return (c.search_order, c.priority)
+
+        candidates = tuple(sorted(candidates, key=prioritize))
 
         # On case sensitive systems, no resolution needed.
         if case_sensitive:
             found_candidate = candidates[0]
         else:
             for candidate in candidates:
-                for fullname, _filename in listDirCached(candidate.found_in):
-                    if fullname == candidate.full_path:
-                        found_candidate = candidate
-                        break
+                if candidate.found_in is None:
+                    found_candidate = candidate
+                else:
+                    for fullname, _filename in listDirCached(candidate.found_in):
+                        if fullname == candidate.full_path:
+                            found_candidate = candidate
+                            break
 
                 if found_candidate:
                     break
@@ -666,14 +740,8 @@ def _findModuleInPath2(package_name, module_name, search_path, logger):
         # Not usable for target architecture.
         raise ImportError
 
-    _reportCandidates(
-        package_name=package_name,
-        module_name=module_name,
-        candidate=found_candidate,
-        candidates=candidates,
-    )
-
     return (
+        found_candidate.found_as,
         found_candidate.full_path,
         "extension" if found_candidate.module_type == "C_EXTENSION" else "py",
     )
@@ -723,6 +791,9 @@ def getPythonUnpackedSearchPath():
 
 
 def getPackageSearchPath(package_name):
+    # TODO: Some branches here are hard coded things that ought to be plugin
+    # decisions, pylint: disable=too-many-branches
+
     if not _main_paths:
         return None
 
@@ -763,9 +834,17 @@ def getPackageSearchPath(package_name):
                 yield extra_path, True
 
         result = []
+
+        # TODO: Make this a plugin decision instead.
+        if package_name == "distutils":
+            setuptools_package_dir = _getSetuptoolsDistutilsPackageDir()
+
+            if setuptools_package_dir is not None:
+                result.append(setuptools_package_dir)
+
         for element in getPackageSearchPath(None):
             for package_dir, force_package in getPackageDirCandidates(element):
-                if isPackageDir(package_dir) or force_package:
+                if force_package or isPackageDir(package_dir):
                     result.append(package_dir)
 
     return OrderedSet(
@@ -788,11 +867,11 @@ def _findModuleInPath(module_name, logger):
         candidate = getLaunchingNuitkaProcessEnvironmentValue("NUITKA_SITE_FILENAME")
 
         if candidate:
-            return candidate, "py"
+            return module_name, candidate, "py"
 
     # Free pass for built-in modules, they need not exist.
     if package_name is None and isBuiltinModuleName(module_name):
-        return None, "built-in"
+        return module_name, None, "built-in"
 
     search_path = getPackageSearchPath(package_name)
 
@@ -803,7 +882,7 @@ def _findModuleInPath(module_name, logger):
         )
 
     try:
-        module_filename, module_kind = _findModuleInPath2(
+        found_module_name, module_filename, module_kind = _findModuleInPath2(
             package_name=package_name,
             module_name=module_name,
             search_path=search_path,
@@ -816,7 +895,7 @@ def _findModuleInPath(module_name, logger):
             module_name if package_name is None else package_name + "." + module_name,
         )
 
-        return None, None
+        return None, None, None
 
     if logger is not None:
         logger.info(
@@ -824,7 +903,7 @@ def _findModuleInPath(module_name, logger):
             % (module_filename, module_kind)
         )
 
-    return module_filename, module_kind
+    return found_module_name, module_filename, module_kind
 
 
 module_search_cache = {}
@@ -879,17 +958,24 @@ def locateModule(module_name, parent_package, level, logger=None):
         kind.
     """
 
+    if not _main_paths:
+        raise NuitkaCodeDeficit(
+            "Error, cannot locate modules before import mechanism is setup."
+        )
+
     if module_name.isMultidistModuleName():
         return locateMultidistModule(module_name)
 
     if _debug_module_finding and logger is None:
         logger = recursion_logger
 
-    module_package, module_filename, module_kind, finding = findModule(
-        module_name=module_name,
-        parent_package=parent_package,
-        level=level,
-        logger=logger,
+    found_module_name, module_package, module_filename, module_kind, finding = (
+        findModule(
+            module_name=module_name,
+            parent_package=parent_package,
+            level=level,
+            logger=logger,
+        )
     )
 
     # Allowing ourselves to be lazy.
@@ -902,8 +988,24 @@ def locateModule(module_name, parent_package, level, logger=None):
     if module_filename is not None:
         module_filename = getNormalizedPath(module_filename)
 
-        module_name, module_kind = getModuleNameAndKindFromFilename(module_filename)
-        module_name = ModuleName.makeModuleNameInPackage(module_name, module_package)
+        if found_module_name == "distutils":
+            _module_name, module_kind = getModuleNameAndKindFromFilename(
+                module_filename
+            )
+            module_name = found_module_name
+        else:
+            module_name, module_kind = getModuleNameAndKindFromFilename(module_filename)
+            module_name = ModuleName.makeModuleNameInPackage(
+                module_name, module_package
+            )
+
+            if _is_debug:
+                assert module_name == found_module_name, (
+                    module_name,
+                    found_module_name,
+                    parent_package,
+                )
+
     elif finding == "not-found":
         if parent_package is not None:
             if not module_name:
@@ -1006,7 +1108,7 @@ def decideModuleSourceRef(filename, module_name, is_main, is_fake, logger):
 # spell-checker: ignore _interpqueues,_lsprof,_multibytecodec,_posixshmem _winapi
 # spell-checker: ignore  _testbuffer _testexternalinspection _testimportmultiple
 # spell-checker: ignore _testinternalcapi _testmultiphase _testsinglephase
-# spell-checker: ignore _xxtestfuzz _xxsubinterpreters imageop
+# spell-checker: ignore _xxtestfuzz _xxsubinterpreters imageop _xxinterpchannels
 _stdlib_module_raises = {
     "_abc": False,
     "__builtin__": False,
@@ -1039,6 +1141,7 @@ _stdlib_module_raises = {
     "_functools": False,
     "_hashlib": False,
     "_heapq": False,
+    "_hmac": False,
     "_hotshot": False,
     "_imp": False,
     "_interpchannels": False,
@@ -1089,6 +1192,7 @@ _stdlib_module_raises = {
     "_tkinter": True,
     "_tokenize": False,
     "_tracemalloc": False,
+    "_types": False,
     "_typing": False,
     "_uuid": False,
     "_warnings": False,
@@ -1097,6 +1201,7 @@ _stdlib_module_raises = {
     "_winreg": False,
     "_xxtestfuzz": False,
     "_xxsubinterpreters": False,
+    "_xxinterpchannels": False,
     "_zoneinfo": False,
     "array": False,
     "atexit": False,
