@@ -69,6 +69,7 @@ from nuitka.PythonVersions import (
 from nuitka.Tracing import plugins_logger
 from nuitka.utils.AppDirs import getAppdirsModule
 from nuitka.utils.Distributions import (
+    getCondaMetaDataVersion,
     getDistributionFromModuleName,
     getDistributionName,
     isDistributionCondaPackage,
@@ -243,11 +244,16 @@ def _getEvaluationContext():
     return _context_dict
 
 
-def _convertVersionToTuple(version_str):
+def _convertVersionToTuple(distribution_name, version_str):
     def numberize(v):
         # For now, we ignore rc/post stuff, hoping it doesn't matter for us.
         return int("".join(d for d in v if d.isdigit()))
 
+    # Anaconda uses placeholders in packages, and different ones come up for
+    # pkg_resources and importlib_metadata
+    if version_str in ("$PKG_VERSION", "-PKG-VERSION") and isAnacondaPython():
+        version_str = getCondaMetaDataVersion(distribution_name)
+
     return tuple(numberize(d) for d in version_str.split("."))
 
 
@@ -286,7 +292,11 @@ def _getPackageVersion(distribution_name):
             else:
                 from importlib_metadata import version
 
-            result = _convertVersionToTuple(version(distribution_name))
+            version_str = version(distribution_name)
+            result = _convertVersionToTuple(
+                distribution_name=distribution_name,
+                version_str=version_str,
+            )
         except ImportError:
             try:
                 from pkg_resources import (
@@ -298,22 +308,26 @@ def _getPackageVersion(distribution_name):
                 result = None
             else:
                 try:
-                    result = _convertVersionToTuple(
-                        get_distribution(distribution_name).version
-                    )
+                    version_str = get_distribution(distribution_name).version
                 except DistributionNotFound:
                     result = None
                 except extern.packaging.version.InvalidVersion:
                     result = None
+                else:
+                    result = _convertVersionToTuple(
+                        distribution_name=distribution_name,
+                        version_str=version_str,
+                    )
 
         if result is None:
             # Fallback if nothing is available, which may happen if no package is installed,
             # but only source code is found.
             try:
                 result = _convertVersionToTuple(
-                    __import__(
+                    distribution_name=distribution_name,
+                    version_str=__import__(
                         _getPackageNameFromDistributionName(distribution_name)
-                    ).__version__
+                    ).__version__,
                 )
             except ImportError:
                 result = None
@@ -438,8 +452,7 @@ class NuitkaPluginBase(getMetaClassBase("Plugin", require_slots=False)):
     and invisible to the user.
 
     Nuitka comes with a number of "standard" plugins to be enabled as needed.
-    What they are can be displayed using "nuitka --plugin-list file.py" (filename
-    required but ignored).
+    What they are can be displayed using "--plugin-list" option.
 
     User plugins may be specified (and implicitly enabled) using their Python
     script pathname.
@@ -1155,6 +1168,10 @@ Unwanted import of '%(unwanted)s' that %(problem)s '%(binding_name)s' encountere
         # Virtual method, pylint: disable=no-self-use,unused-argument
         return None
 
+    def decideRecompileExtensionModules(self, module_name):
+        # Virtual method, pylint: disable=no-self-use,unused-argument
+        return None
+
     def decideAnnotations(self, module_name):
         # Virtual method, pylint: disable=no-self-use,unused-argument
         return None
@@ -1270,14 +1287,19 @@ Unwanted import of '%(unwanted)s' that %(problem)s '%(binding_name)s' encountere
         # Virtual method, pylint: disable=no-self-use
         return None
 
-    def warnUnusedPlugin(self, message):
+    def warnUnusedPlugin(self, message, key=None):
         """An inactive plugin may issue a warning if it believes this may be wrong.
 
         Returns:
             None
         """
-        if self.plugin_name not in _warned_unused_plugins:
-            _warned_unused_plugins.add(self.plugin_name)
+        if key is None:
+            key = self.plugin_name, message
+        else:
+            key = self.plugin_name, key
+
+        if key not in _warned_unused_plugins:
+            _warned_unused_plugins.add(key)
 
             plugins_logger.warning(
                 "Use '--enable-plugin=%s' for: %s" % (self.plugin_name, message)
@@ -1322,7 +1344,7 @@ Unwanted import of '%(unwanted)s' that %(problem)s '%(binding_name)s' encountere
             return self._runtime_information_cache[info_name]
 
         keys = []
-        query_codes = []
+        query_codes = ['print("-*-" * 15)']
 
         for key, value_expression in values:
             keys.append(key)
@@ -1362,8 +1384,13 @@ except Exception as e:
         env = dict(os.environ)
         env["PYTHONIOENCODING"] = "utf8"
 
+        command = [sys.executable, "-c", cmd]
+
+        if isMacOS():
+            command = ["arch", "-" + Options.getMacOSTargetArch()] + command
+
         try:
-            feedback = check_output([sys.executable, "-c", cmd], env=env)
+            feedback = check_output(command, env=env)
         except NuitkaCalledProcessError as e:
             if e.returncode == 38:
                 self.warning(
@@ -1397,6 +1424,9 @@ except Exception as e:
         # Ignore Windows newlines difference.
         feedback = [line.strip() for line in feedback.splitlines()]
 
+        while feedback and not feedback[0] != "-*-" * 15:
+            feedback.pop(0)
+
         if feedback.count("-" * 27) != len(keys):
             self.sysexit(
                 "Error, mismatch in output retrieving %r information." % info_name
@@ -1428,7 +1458,9 @@ except Exception as e:
             values=(("key", value),),
         ).key
 
-    def onFunctionBodyParsing(self, module_name, function_name, body):
+    def onFunctionBodyParsing(
+        self, module_name, function_qualname, function_name, body
+    ):
         """Provide a different function body for the function of that module.
 
         Should return a boolean, indicating if any actual change was done.
@@ -1845,6 +1877,33 @@ Error, expression '%s' for module '%s' did not evaluate to 'tuple[str]' or 'list
         )
 
 
+class NuitkaDetectorPluginBase(NuitkaPluginBase):
+    """Base class for detectors."""
+
+    @classmethod
+    def isRelevant(cls):
+        """Always relevant by default."""
+        return True
+
+    # Needs to be another plugin.
+    detector_for = None
+
+
+class NuitkaNamespaceDetectorPluginBase(NuitkaDetectorPluginBase):
+    """Base class for a detector just waiting for a namespace to appear."""
+
+    # TODO: This actually would be done best when the module set is complete, as
+    # in theory, this module using it could become unused again.
+    def onModuleDiscovered(self, module):
+        """This method checks whether detected plugin is required for a module."""
+        for detector_namespace in self.detector_namespaces:
+            if module.getFullName().hasNamespace(detector_namespace):
+                self.warnUnusedPlugin("Missing '%s' support." % detector_namespace)
+
+    # Should be a tuple
+    detector_namespaces = ()
+
+
 class TagContext(dict):
     def __init__(self, logger, full_name, config_name):
         dict.__init__(self)
